#ifndef PLOTTER_H
#define PLOTTER_H

// standard
#include <cstdlib>
#include <cmath>
#include <iostream>
// stl
#include <vector>
#include <string>
#include <map>
// gsl
#include <gsl/gsl_integration.h>
#include <gsl/gsl_monte.h>
#include <gsl/gsl_monte_plain.h>
#include <gsl/gsl_monte_vegas.h>
#include <gsl/gsl_rng.h>
// pdf
//#include <TMDlib.h>
#include <LHAPDF/LHAPDF.h>

namespace TMDplotter
{

/*
 * A general Curve object that stores information about and provides functions to plot a PDF curve
 * Most parts are virtual and only implemented in derived classes
 */
class Curve
{
public:
    // constructor and destructor
    ~Curve();
    Curve(const std::string, const std::string, const std::string, double, std::map<std::string, short>);
    virtual void init() =0;
    void setKtIntegrationLimits (double, double);
    // This function returns the y value of a non-integrated (= differentiated) PDF to given values for x and kt
    virtual double getUnintegrated(double, double, double) =0;
    std::vector<double> getUnintegrated(std::vector<double>,             double,             double );
    std::vector<double> getUnintegrated(            double, std::vector<double>,             double );
    std::vector<double> getUnintegrated(            double,             double, std::vector<double>);
    std::vector<double> getUnintegratedAllMembers (double, double, double);
    template<typename T, typename U, typename V> std::vector<std::vector<double> > getUnintegratedAllMembers (T, U, V);
    //std::pair<double, double> getUnintegratedEnvelope(double, double, double);
    //std::pair<double, double> getUnintegratedBand(double, double, double);
    template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > getUnintegratedEnvelope(T, U, V);
    template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > getUnintegratedBand(T, U, V);
    template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > getUnintegratedMC(T, U, V);
    // This function returns the y value of an integrated PDF to a given x-value
    virtual double getIntegrated(double, double) =0;
    std::vector<double> getIntegrated(std::vector<double>,             double );
    std::vector<double> getIntegrated(            double, std::vector<double>);
    std::vector<double> getIntegratedAllMembers (double, double);
    template<typename T, typename U> std::vector<std::vector<double> > getIntegratedAllMembers (T, U);
    //std::pair<double, double> getIntegratedEnvelope(double, double);
    //std::pair<double, double> getIntegratedBand(double, double);
    template<typename T, typename U> std::pair<std::vector<double>,std::vector<double> > getIntegratedEnvelope(T, U);
    template<typename T, typename U> std::pair<std::vector<double>,std::vector<double> > getIntegratedBand(T, U);
    template<typename T, typename U> std::pair<std::vector<double>,std::vector<double> > getIntegratedMC(T, U);
    // get methods for Legend and ScaleFactor
    long double getKtMin ();
    long double getKtMax ();
    std::string getEvolution();
    virtual std::string getParton();
    std::string getSet();
    std::string getLegend();
    long double getScalefactor();
    // static function to set the current program name which is printed in outputs
    static void set_program_name(std::string);
    static std::string program_name;
    bool test; // if this bool is set true, the object has to be treated as test object (alternative calculation methods)
protected:
    Curve(const std::string, const std::string, double);
    const std::string evolution, parton, name; // carries the parton name (for luminosities: a combination of both parton names) and the name of the PDF set
    std::map<std::string, short> partons;
    int parton_id, // carries the parton id (not set for luminosities)
        imember, // defines which member/subset is to be used
        nmembers;
    double scaleFactor, xbar; // carries scaleFactor and xbar
    double ktmin, ktmax;
    unsigned ALLOCATEDSPACE; // size of the workspace for integration (set to 1000)
    gsl_integration_workspace * myworkspace; // workspace for integrating over k_t
    gsl_function integrationFunction; // integration function for integrating over k_t
};

/*
 * A Curve object that stores information about and provides functions to plot a TMD PDF curve
 * Implements all virtual functions of class Curve, tailored to TMD PDFs
 * Detailed information about functions in .cc-file or in base class Curve
 */
class CurveTMD : public Curve
{
    unsigned iset_nominal;
public:
    // constructors and destructor
    ~CurveTMD();
    CurveTMD(std::string, std::string, double);
    // functions
    void init();
    double getUnintegrated(double, double, double);
    double getIntegrated(double, double);
    double rescaled_uTMD_integrand(double& r, void * params);
};

/*
 * A Curve object that stores information about and provides functions to plot a DGLAP PDF curve
 * Implements all virtual functions of class Curve, tailored to DGLAP PDFs
 * Detailed information about functions in .cc-file or in base class Curve
 */
class CurveDGLAP : public Curve
{
public:
    // constructors and destructor
    ~CurveDGLAP();
    CurveDGLAP(std::string, std::string, double);
    // functions
    void init();
    double getUnintegrated(double, double, double);
    double getIntegrated(double, double);
    double testIntegrand(double&, void *);
protected:
    std::vector<LHAPDF::PDF*> lha_pdf;
private:
    const double derivative_step;
};

/*
 * A Curve object that stores information about and provides functions to plot a Luminosity curve
 * Implements all virtual functions of class Curve, tailored to Luminosities
 * Detailed information about functions in .cc-file or in base class Curve
 */
class Luminosity : public Curve
{
public:
    // constructors and destructor
    ~Luminosity();
    Luminosity(std::string, std::string, std::string, std::string, double, double = 0);
    // functions
    void init();
    double getUnintegrated(double, double, double);
    double getIntegrated(double, double);
    double integrand(double&, void *); // integrand for integrated luminosity
    double testIntegrand(double&, void *); // duplicate integrand, used in testmode (alternative calculation of integrated luminosity)
    double integrand(double&, double&, double&, void *); // integrand for unintegrated luminosity
    std::string getParton();
private:
    const double s;
    TMDplotter::Curve * firstPDF; // PDF of first parton
    TMDplotter::Curve * secondPDF; // PDF of second parton
    const double xmax, symmetryfactor; // xmax: upper limit of integration, symmetryfactor: 1/(1-delta(ij)) for partons i and j
    gsl_rng * r; // random number generator for MC integration
    gsl_monte_vegas_state * myworkstate; // workstate for Monte Carlo integration over xone, kt and phi (unint. luminosities)
    gsl_monte_function monteIntegrationFunction; // integration function for MC integration
    const int MCsamples; // number of Monte Carlo samples
}; // end class Luminosity

/*
 *
 * Helper functions
 *
 */
template<typename T> struct integration_params
{
    double x, mu, qt;
    T * myObject;
}; // structure to pass integration parameters to wrappers and integrands
std::vector<double> min (std::vector<double>, std::vector<double>);
std::vector<double> max (std::vector<double>, std::vector<double>);
std::vector<double> hypot (std::vector<double>, std::vector<double>);
std::vector<double> add (std::vector<double>, std::vector<double>);
std::vector<double> subtract (std::vector<double>, std::vector<double>);
double differentiate (TMDplotter::Curve*, double, double, double); // derivation function for DGLAP
double Lumi_integrand_wrapper(double, void *); // wrapper for integrating luminosities over x
double unintLumi_integrand_wrapper(double*, size_t, void *); // wrapper for mc-integrating 'unintegrated' luminosities
double TMD_integrand_wrapper(double, void *); // wrapper for integrating TMD TMDs over kt
double DGLAP_integrand_wrapper(double, void *);

template<typename T, typename U, typename V> std::vector<std::vector<double> > Curve::getUnintegratedAllMembers (T x, U kt, V mu)
{
    std::vector<std::vector<double> > members;
    for (imember = 0; imember < nmembers ; imember++)
        members.push_back(getUnintegrated(x, kt, mu));
    imember = 0;
    return members;
}
template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getUnintegratedEnvelope (T x, U kt, V mu)
{
    std::vector<std::vector<double> > members = Curve::getUnintegratedAllMembers (x, kt, mu);
    std::vector<double> lower_unc = members[0],
                        upper_unc = members[0];
    for (imember = 1; imember < nmembers ; imember++)
    {
        upper_unc = TMDplotter::max(upper_unc, members[imember]);
        lower_unc = TMDplotter::min(lower_unc, members[imember]);
    }
    imember = 0;
    return {upper_unc, lower_unc};
}
template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getUnintegratedBand (T x, U kt, V mu)
{
    std::vector<std::vector<double> > members = Curve::getUnintegratedAllMembers (x, kt, mu);
    std::vector<double> lower_unc(members[0].size(),0),
        upper_unc(members[0].size(),0);
    for (imember = 1; imember < nmembers ; imember++)
    {
        std::vector<double> Dx = TMDplotter::subtract(members[imember],members[0]);
        //if (members[imember][0] > members[0][0]) upper_unc = TMDplotter::hypot(upper_unc, Dx);
        //else
        //if (members[imember][0] < members[0][0]) lower_unc = TMDplotter::hypot(lower_unc, Dx);
        if (imember % 2 == 0)  lower_unc = TMDplotter::hypot(lower_unc, Dx);
        else
            upper_unc = TMDplotter::hypot(lower_unc, Dx);
    }
    //cout << "  getUnitegratedAll" << endl;
    upper_unc =      TMDplotter::add(members[0],upper_unc);
    lower_unc = TMDplotter::subtract(members[0],lower_unc);

    for (size_t i = 0 ; i < members[0].size() ; i++)
    {
        //  std::cout << " central " << members[0][i] <<  " unc  " <<  upper_unc[i] << " " << lower_unc[i] << " " << i << std::endl;
    }

    imember = 0;
    return {upper_unc, lower_unc};
}
template<typename T, typename U, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getUnintegratedMC (T x, U kt, V mu)
{
    std::vector<std::vector<double> > members = Curve::getUnintegratedAllMembers (x, kt, mu);
    std::vector<double> lower_unc(members[0].size(),0),
        upper_unc(members[0].size(),0);
    // std::cout << " in getUnintegratedMC " << nmembers << std::endl;
    std::vector<double> rms2(members[0].size(),0),rms(members[0].size(),0), Dx(members[0].size(),0);
    for (size_t i = 0 ; i < members[0].size() ; i++)
    {
        rms2[i] = 0;
    }
    for (imember = 1; imember < nmembers ; imember++)
    {
        //std::cout << " members " << imember << std::endl;
        for (size_t i = 0 ; i < members[0].size() ; i++)
        {
            Dx[i]  = (members[imember][i] - members[0][i])*(members[imember][i] - members[0][i]);
            rms2[i] = rms2[i] + Dx[i] ;
            //std::cout << " rms2 " << rms2[i] << " i= "<< i << std::endl;
        }
    }
    for (size_t i = 0 ; i < members[0].size() ; i++)
    {
        //std::cout << " rms2 " << rms2[i] << " i= "<< i << std::endl;
        rms[i] = 0;
        if(nmembers > 1) rms[i] = std::sqrt(rms2[i]/(nmembers - 1));
    }
    // std::cout << "before end  " << imember << std::endl;
    upper_unc = TMDplotter::add(members[0],rms);
    lower_unc = TMDplotter::subtract(members[0],rms);
    /*
            for (size_t i = 0 ; i < members[0].size() ; i++)
            {
               std::cout << " central " << members[0][i] <<  " unc  " <<  upper_unc[i] << " " << lower_unc[i] << " " << i << std::endl;
            }
    */
    imember = 0;
    return {upper_unc, lower_unc};
}

template<typename T, typename V> std::vector<std::vector<double> > Curve::getIntegratedAllMembers (T x, V mu)
{
    std::vector<std::vector<double> > members;
    for (imember = 0; imember < nmembers ; imember++)
        members.push_back(getIntegrated(x, mu));
    imember = 0;
    return members;
}
template<typename T, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getIntegratedEnvelope (T x, V mu)
{
    std::vector<std::vector<double> > members = Curve::getIntegratedAllMembers (x, mu);
    std::vector<double> lower_unc = members[0],
                        upper_unc = members[0];
    for (imember = 1; imember < nmembers ; imember++)
    {
        upper_unc = TMDplotter::max(upper_unc, members[imember]);
        lower_unc = TMDplotter::min(lower_unc, members[imember]);
    }
    imember = 0;
    return {upper_unc, lower_unc};
}
template<typename T, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getIntegratedBand (T x, V mu)
{
    std::vector<std::vector<double> > members = Curve::getIntegratedAllMembers (x, mu);
    std::vector<double> lower_unc(members[0].size(),0),
        upper_unc(members[0].size(),0);
    for (imember = 1; imember < nmembers ; imember++)
    {
        std::vector<double> Dx = TMDplotter::subtract(members[imember],members[0]);
        if (members[imember][0] > members[0][0]) upper_unc = TMDplotter::hypot(upper_unc, Dx);
        else if (members[imember][0] < members[0][0]) lower_unc = TMDplotter::hypot(lower_unc, Dx);
    }
    upper_unc =      TMDplotter::add(members[0],upper_unc);
    lower_unc = TMDplotter::subtract(members[0],lower_unc);
    imember = 0;
    return {upper_unc, lower_unc};
}
template<typename T, typename V> std::pair<std::vector<double>, std::vector<double> > Curve::getIntegratedMC (T x, V mu)
{
    std::vector<std::vector<double> > members = Curve::getIntegratedAllMembers (x, mu);
    std::vector<double> lower_unc(members[0].size(),0),
        upper_unc(members[0].size(),0);
    // std::cout << " in getIntegratedMC " << nmembers << std::endl;

    std::vector<double> rms2(members[0].size(),0),rms(members[0].size(),0), Dx(members[0].size(),0);
    for (size_t i = 0 ; i < members[0].size() ; i++)
    {
        rms2[i] = 0;
    }
    for (imember = 1; imember < nmembers ; imember++)
    {
        //std::cout << " members " << imember << std::endl;
        for (size_t i = 0 ; i < members[0].size() ; i++)
        {
            Dx[i]  = (members[imember][i] - members[0][i])*(members[imember][i] - members[0][i]);
            rms2[i] = rms2[i] + Dx[i] ;
            //std::cout << " rms2 " << rms2[i] << " i= "<< i << std::endl;
        }
    }
    for (size_t i = 0 ; i < members[0].size() ; i++)
    {
        //std::cout << " rms2 " << rms2[i] << " i= "<< i << std::endl;
        rms[i] = 0;
        if(nmembers > 1) rms[i] = std::sqrt(rms2[i]/(nmembers - 1));
    }
    // std::cout << "before end  " << imember << std::endl;
    upper_unc = TMDplotter::add(members[0],rms);
    lower_unc = TMDplotter::subtract(members[0],rms);
    /*
            for (size_t i = 0 ; i < members[0].size() ; i++)
            {
               std::cout << " central " << members[0][i] <<  " unc  " <<  upper_unc[i] << " " << lower_unc[i] << " " << i << std::endl;
            }
    */
    imember = 0;
    return {upper_unc, lower_unc};

}
}

#endif
