
// standard
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <fstream>
#include <time.h>
#include <vector>
// gsl
#include <gsl/gsl_integration.h>

// pdf
#include <tmdlib/TMDlib.h>
#include "TMDlibPDFpath.h"
#include <LHAPDF/LHAPDF.h>

// Root includes
#include <TH2.h>
#include <TCanvas.h>
#include <TString.h>
#include <TLegend.h>
#include <TText.h>
#include <TLatex.h>
#include <TGraph.h>
#include <TGraphAsymmErrors.h>
//#include <TASImage.h>

// Custom classes
#include "Curve.h"
#include "tmdexception.h"
#include <cmath>

using namespace std;

namespace TMDplotter
{
/*
 * Here, some helper methods for the main are implemented
 */

// This method sets up the vector of x resp. kt  values
// stepping mode: 0 for linear, 1 for logarithmic stepping, (possibly to be continued?)
// min, max: limits for the values of x resp. kt
// nsteps: number of nsteps in interval [min, max]
vector<double> stepping(bool logx, double xmin, double xmax, int nsteps, const string program_name)
{
    double x = xmin;
    vector<double> steps;
    enum Stepmode {LINEAR, LOGARITHMIC};
    switch(logx)
    {
    case LINEAR:
    {
        const double steplength = (xmax-xmin)/double(nsteps);
        for(int i = 0; i<=nsteps; i++)
        {
            steps.push_back(x);
            x+=steplength;
        }
        // cout << program_name << "\tDoing " << nsteps << " linear nsteps (step length = " << steplength << ")" << endl;
    }
    break;
    case LOGARITHMIC:
    {
        const double steplength = (log10(xmax/xmin))/double(nsteps);
        for(int i = 0; i<=nsteps; i++)
        {
            steps.push_back(x);
            x*=pow(10,steplength);
        }
        // cout << program_name << "\tDoing " << nsteps << " logarithmic nsteps (logstep = " << steplength << ")" << endl;
    }
    break;
    }
    return steps;
}

// This method gets an argument string in format "$variablename=$variablevalue" and returns a pair<string, string>($name,$value)
// Argument: The argument string to be splitted
pair<string, string> getArgument(string argument)
{
    stringstream ss(argument);
    string name;
    string content;
    getline(ss, name, '=');
    getline(ss, content);
    return pair<string, string>(name,content);
}

string remove_trailing_zeros (string number)
{
    //if (number.find('.') != number.end()) // C++11
    //    while (number.back() == '0')
    //        number.pop_back();
    if (number.find('.') != string::npos)
    {
        while (number[number.size()-1] == '0' || number[number.size()-1] == '.')
        {
            if (number[number.size()-1] == '.')
            {
                number.erase(number.end()-1,number.end());
                break;
            }
            number.erase(number.end()-1,number.end());
        }
    }
    return number;
}

/*
 * End of helper methods
 */
}
//END OF NAMESPACE TMDplotter

/*
 *
 * Begin of main HERE
 *
 */
int main (int argc, char * argv[])
{
    int start = time(NULL);
    using namespace TMDplotter;
    const string program_name = argv[0];
    // check if help is asked and if enough arguments are given
    if (argc==1)
    {
        // cout << program_name << "\tNo arguments given" << endl; // only one option
        cout << " TMDplotter: " << "\tNo arguments given" << endl; // only one option
        return EXIT_FAILURE;
    }
    if (!(strcmp(argv[1], "-h") && strcmp(argv[1], "help") && strcmp(argv[1], "-help") && strcmp(argv[1], "--help"))) // help case
    {
        //cout << program_name << " [mode] type\n"
        cout << " TMDplotter: " << " [mode] type\n"
             << "\t[-u]\n"
             << "\t[-a]\n"
             << "\t[-b]\n"
             << "\t[-g] [global options] \n"
             << "\t-c curve options \n"
             << "\t[-c curve options] \n"
             << "\t[...]\n"
             << "\n"
             << "\tpossible global-only options: \n"
             << "\t\txmin= xmax=\n"
             << "\t\tktmin= ktmax= \n"
             << "\t\tmumin= mumax=\n"
             << "\t\tnsteps=\n"
             //<< "\t\ts= (this option will also activate plotting over sqrt(s) for luminosity)\n"
             << "\t\tdirectory= filename= filetype= \n"
             << "\n"
             << "\tpossible global AND curve options: \n"
             << "\t\tx= kt= mu= sf=\n"
             << "\t\tevolution= name= parton= [firstparton=] [secondparton=] \n"
             << "For further explanation, see the readme.txt file or documentation" << endl;
        return EXIT_SUCCESS;
    }
    //// cout the entries
    //clog << "argc = " << argc;
    //for (short iarg = 0 ; iarg < argc ; iarg++)
    //	clog << "\nargv[" << iarg << "] = " << argv[iarg];
    //clog << endl;

    // first, retrieve type (and, if given, mode)
    // therefore, loop over arguments until first argument with "-" appears. The option before that is the type (and before that, if any, is the mode)
    string modestring, typestring;
    short narg = 1;
    cout << " TMDplotter Messages: " << endl;

    while ((narg<argc) && !(argv[narg][0] == '-'))
        narg++; // just increase narg until first "-xxx" option appears
    if(argv[narg][0] == '-')
    {
        if(narg<2) // narg=1, no type and no mode is specified --> Error!
        {
            //cout << program_name << "\tPlease specify type first! Possible types are: \ttmdx\ttmdkt\ttmdmu\tpdfx\tpdfmu" << endl;
            cout << " TMDplotter: " << "\tPlease specify type first! Possible types are: \ttmdx\ttmdkt\ttmdmu\tpdfx\tpdfmu" << endl;
            return EXIT_FAILURE;
        }
        else if (narg==2) // only type is specified, mode is default
        {
            modestring = "default";
            typestring=argv[1];
        }
        else if (narg==3) // mode and type are specified
        {
            modestring=argv[1];
            typestring=argv[2];
        }
        else // narg>3, more than mode and type is specified --> invalid!
        {
            //cout << program_name << "\tInvalid option:" << argv[narg-1] << endl;
            cout << " TMDplotter: " << "\tInvalid option:" << argv[narg-1] << endl;
            return EXIT_FAILURE;
        }
    }
    else // no "-xxx" given --> invalid!
    {
        //cout << program_name << "\tPlease specify curves and global options!" << endl;
        cout << " TMDplotter: " << "\tPlease specify curves and global options!" << endl;
        return EXIT_FAILURE;
    }

    // init mode and type

    enum Mode
    {
        DEFAULT,  // 0
        LUMINOSITY,  // 1
    };
    enum Type
    {
        TMDX,  // 0
        TMDKT,  // 1
        TMDMU,  // 2
        PDFX,  // 3
        PDFMU,  // 4
    };

    short mode;
    if(modestring == "default")  // default mode
        mode=DEFAULT;
    else if(modestring == "luminosity")  // lumi mode
        mode=LUMINOSITY;
    else  // unknown mode
    {
        //cout << program_name << "\tImpossible mode: " << modestring << endl;
        cout << " TMDplotter: " << "\tImpossible mode: " << modestring << endl;
        return EXIT_FAILURE;
    }
    short type;
    if(typestring == "tmdx")  // unintegrated function of x
        type=TMDX;
    else if(typestring == "tmdkt")  // unintegrated function of kt
        type=TMDKT;
    else if(typestring == "tmdmu")  // unintegrated function of mu
        type=TMDMU;
    else if(typestring == "pdfx")  // integrated function of x
        type=PDFX;
    else if(typestring == "pdfmu")  // integrated function of mu
        type=PDFMU;
    else // unknown type
    {
        cout << " TMDplotter: " << "\tImpossible type: " << typestring << endl;
        return EXIT_FAILURE;
    }

    //set up maps for the input values
    vector<map<string,string> > curveparams; // this will contain the distinct parameters for the curves
    map<string,string> globalparams; // this will contain the global parameters

    bool test = false, central_value = true, band = false, uncertainty = false, mc = false, allmembers = false;
    for(short i = narg ; i < argc ; i++)
    {
        string current_argument = argv[i];
        test          =    test          || current_argument == "-t" || current_argument == "--test"             ;
        central_value = !(!central_value || current_argument == "-n" || current_argument == "--no-central-value");
        uncertainty   =    uncertainty   || current_argument == "-u" || current_argument == "--uncertainty"      ;
        band          =    band          || current_argument == "-b" || current_argument == "--band"             ;
        mc            =    mc            || current_argument == "-mc" || current_argument == "--monte-carlo"             ;
        allmembers    =    allmembers    || current_argument == "-a" || current_argument == "--allmembers"       ;
    }

    if (test)
        cout << " TMDplotter: " << "\tTest mode activated" << endl;
    if (!central_value)
        cout << " TMDplotter: " << "\tcentral value will not be plotted" << endl;
    if (uncertainty)
        cout << " TMDplotter: " << "\tenvelope mode activated" << endl;
    if (band)
        cout << " TMDplotter: " << "\tband mode activated" << endl;
    if (mc)
        cout << " TMDplotter: " << "\tMonte Carlo replica mode activated" << endl;
    if (allmembers)
        cout << " TMDplotter: " << "\tallmembers mode activated" << endl;

    // initialize GLOBAL parameters according to config.ini
    fstream configfile;

    // cout << " pdfpath " << pdfpath << endl;
    configfile.open(pdfpath+"config_tmdplotter.ini", fstream::in);
    // configfile.open("/Library/WebServer/Documents/NewTMDplotter/include/root/config.ini", fstream::in); // TODO: unhardcode this path!!!
    while (!configfile.eof())
    {
        string line;
        getline(configfile,line);
        if(line.length()!=0 && line[0]!=';')
        {
            pair<string,string> valuepair=getArgument(line);
            globalparams[valuepair.first]=valuepair.second;
        }
    }
    configfile.close();

    // loop twice over remaining cmdline arguments and get parameters, save them to map(s) - one time global ones and one time curve specific ones
    for(short i=narg; i<(argc-1); i++) if( !( strcmp(argv[i], "-g") && strcmp(argv[i], "--global") ) ) // global parameters following
        {
            while((i+1) < argc && !(argv[i+1][0] == '-')) // !(next option begins or end of argv)
            {
                i++;
                pair<string,string> valuepair=getArgument(argv[i]);
                globalparams[valuepair.first]=valuepair.second;
            }
        }

    for(int i=narg; i<(argc-1); i++) if( !( strcmp(argv[i], "-c") && strcmp(argv[i], "--curve") ) ) // parameters for a single curve following
        {
            map<string,string> currentCurveParas=globalparams;
            while((i+1) < argc && argv[i+1][0] != '-') // !(next option begins or end of argv)
            {
                i++;
                pair<string,string> valuepair=getArgument(argv[i]);
                currentCurveParas[valuepair.first]=valuepair.second;
            }
            curveparams.push_back(currentCurveParas);
        }

    /*
     *
     * Retrieval done, now set parameters
     *
     */


    // set global parameters first

    const bool draw_ratio=stoi(globalparams["ratio"]); // draw a ratio?
    const bool logx=stoi(globalparams["logx"]); // stepping mode (log/lin)
    const bool logy=stoi(globalparams["logy"]); // stepping mode (log/lin)
    const bool ratiology=draw_ratio ? stoi(globalparams["ratiology"]) : logy; // stepping mode (log/lin)
    const bool overs= (globalparams.find("overs") != globalparams.end() && mode == LUMINOSITY) && stoi(globalparams["overs"]);
    const unsigned int nsteps=max(2,stoi(globalparams["nsteps"])); // number of nsteps
    const double xmin=stod(globalparams["xmin"]),
                 xmax=stod(globalparams["xmax"]),
                 ktmin=stod(globalparams["ktmin"]),
                 ktmax=stod(globalparams["ktmax"]),
                 mumin=stod(globalparams["mumin"]),
                 mumax=stod(globalparams["mumax"]),
                 ymin=stod(globalparams["ymin"]),
                 ymax=stod(globalparams["ymax"]),
                 ratioymin=stod(globalparams["ratioymin"]),
                 ratioymax=stod(globalparams["ratioymax"]);
    const string directory = globalparams["directory"] + (globalparams["directory"][globalparams["directory"].size()-1] == '/' ? "" : "/"),  // the last '/' is added only if not already there
                 filename  = globalparams["filename"],  // ex: foo, bar, ...
                 filetype  = globalparams["filetype"], // ex: pdf, png, eps, ... (any root format)
                 output_name = directory + filename + "." + filetype;

    // control output of global parameters
    /*
    cout << program_name << "\txmin = " << xmin << ", xmax = " << xmax << '\n'
    	 << program_name << "\tktmin = " << ktmin << ", ktmax = " << ktmax << '\n'
    	 << program_name << "\tmumin = " << mumin << ", mumax = " << mumax << '\n'
    	 << program_name << "\tymin = " << ymin << ", ymax = " << ymax << '\n'
    	 << program_name << "\tratioymin = " << ratioymin << ", ratioymax = " << ratioymax << '\n'
    	 << program_name << "\tnsteps = " << nsteps << '\n'
    	 << program_name << "\tovers = " << overs << '\n'
    	 << program_name << "\toutput: " << output_name << endl;
         */
    // set program name in class Curve, used for print on screen
    Curve::set_program_name(program_name);
    // Vectors with curve pointers and values at which curves are supposed to be plotted
    vector<long double> xBvalues,  // fixed x values for all curves
           ktvalues, // fixed kt values for all curves
           muvalues,  // fixed mu values for all curves
           sqrtsvalues  ; // fixed mu values for all curves
    vector<Curve*> curves; // vector of all Curve objects
    bool isTestCurve = false; // in test mode, every Curve object gets a duplicate marked as test Curve
    // Retrieving the different PDFs to plot
    for(unsigned int i=0; i<curveparams.size(); i++)
    {
        Curve * current_curve;
        // add the current pdf to the list of PDFs to plot
        // setting of local parameters
        const double current_SF       = stod(curveparams[i]["sf"]),
                     current_fixedx   = stod(curveparams[i]["x"]),
                     current_fixedkt  = stod(curveparams[i]["kt"]),
                     current_fixedmu  = stod(curveparams[i]["mu"]),
                     current_intktmin = curveparams[i].find("intktmin") != curveparams[i].end() ? stod(curveparams[i]["intktmin"]) : ktmin,
                     current_intktmax = curveparams[i].find("intktmax") != curveparams[i].end() ? stod(curveparams[i]["intktmax"]) : current_fixedmu,
                     current_sqrts    = overs ? stod(curveparams[i]["sqrts"]) : 0;
        const string current_evolution    = curveparams[i]["evolution"],
                     current_name         = curveparams[i]["name"],
                     current_parton       = curveparams[i]["parton"],
                     current_firstparton  = curveparams[i][curveparams[i].find("firstparton")  != curveparams[i].end() ? "firstparton"  : "parton"],
                     current_secondparton = curveparams[i][curveparams[i].find("secondparton") != curveparams[i].end() ? "secondparton" : "parton"];
        // control output of local parameters
        /*
        cout << program_name << "\tCurve# = " << i << '\n'
        		 << program_name << "\tevolution = " << current_evolution << " && name = " << current_name << '\n'
        		 << program_name << "\tparton = " << current_parton << '\n'
        		 << program_name << "\t(firstparton = " << current_firstparton << " && secondparton = " << current_secondparton << ")" << '\n'
        		 << program_name << "\tscale factor = " << current_SF << '\n'
        		 << program_name << "\tx = " << current_fixedx << " && kt = " << current_fixedkt << " && mu = " << current_fixedmu << '\n'
                   << program_name << "\tintktmin = " << current_intktmin << " && intktmax = " << current_intktmax << '\n';
              if (mode == LUMINOSITY)
                   cout << program_name << "\tsqrt(s) = " << current_sqrts << '\n';
              cout << endl;
              */

        try
        {
            switch(mode)
            {
            case DEFAULT: // create TMD or DGLAP Curve objects
                if (current_evolution == "DGLAP")
                    current_curve = new CurveDGLAP(current_parton, current_name, current_SF); // Parton, name, scale factor
                else if (current_evolution == "TMD" )
                    current_curve = new CurveTMD(current_parton, current_name, current_SF); // Parton, name, scale factor
                else
                    throw Unknownevolution(current_evolution, __LINE__);
                break;
            case LUMINOSITY: // create Luminosity Curve objects
                current_curve = new Luminosity(current_evolution, current_firstparton, current_secondparton, current_name, current_SF, current_sqrts); // evolution, 1st parton, 2nd parton, name, scale factor TODO: check sqrt(s) and add uncertainty
                break;
            default:
                throw Notimplemented("*unkwown mode*", __LINE__);
            } // end switch
            current_curve->setKtIntegrationLimits(current_intktmin, current_intktmax);
            // push-back values for x, kt and mu
            xBvalues   .push_back(current_fixedx );
            ktvalues   .push_back(current_fixedkt);
            muvalues   .push_back(current_fixedmu);
            curves     .push_back(current_curve);
            sqrtsvalues.push_back(current_sqrts);
            if(test)
            {   // If test mode is activated, every curve will be duplicated and the duplicate will be marked as a 'test curve'
                current_curve->test = isTestCurve; // mark as test curve
                if(!isTestCurve) i--; // if duplicate wasn't created yet, create it --> decrease i to repeat step
                isTestCurve = !(isTestCurve); // if this was a test duplicate, next one is not and vice versa
            }
        }
        catch(Notimplemented& n)
        {   // TMDexception for unimplemented functions
            cout << program_name << '\t' << n.what() << " Program will end now." << endl;
            return EXIT_FAILURE;
        }
        catch(Unknownparton& p)
        {   // TMDexception for unknown parton
            cout << program_name << '\t' << p.what() << endl;
            throw;
        }
        catch(Unknownevolution& e)
        {   // TMDexception for unknown evolution equation
            cout << program_name << '\t' << e.what() << endl;
            throw;
        }
        catch(LHAPDF::ReadError l)
        {   // LHAPDF exception for invalid PDF
            cout << program_name << '\t' << l.what() << endl;
            throw;
        }
        catch (...)
        {
            cout << program_name << "\tIgnoring current entry." << endl;
            continue;
        }
    } // end loop over curveparams

    const unsigned ncurves = curves.size();
    // cout << program_name << '\t' << ncurves << " curve(s) will be produced:" << endl;
//    for (size_t i = 0 ; i < ncurves ; i++)
//          cout << program_name << '\t' << curves[i]->getParton() << '\t' << curves[i]->getSet() << endl;
    // stepping
    vector<double> X; // remember that "x" means here X-axis, regardless of the variable (kt, mu, or xB)
    switch(type)
    {
    case PDFX:
    case TMDX:
        X = stepping(logx,  xmin,  xmax, nsteps, program_name);
        break;
    case TMDKT:
        X = stepping(logx, ktmin, ktmax, nsteps, program_name);
        break;
    case PDFMU:
    case TMDMU:
        X = stepping(logx, mumin, mumax, nsteps, program_name);
    }
    vector<vector<double> > v_Y, v_Yup, v_Ylow, v_Yup2, v_Ylow2; // This will contain the points to plot
    vector<vector<vector<double> > > v_all;
    try
    {
        for(unsigned int i=0; i<ncurves; i++)
        {
#ifdef DEBUG
            cout << " here call curves:init" << endl;
#endif
            curves[i]->init();
            vector<double> Y, Yup, Ylow;
            pair<vector<double>,vector<double>> unc, ban;
            pair<vector<double>,vector<double>> unc1, mcban;
            vector<vector<double> > all;
            try
            {
                switch(type)
                {
                case TMDX:
                    v_Y.push_back(curves[i]->getUnintegrated(X, ktvalues[i], muvalues[i]));
                    if (allmembers)
                        all = curves[i]->getUnintegratedAllMembers(X, ktvalues[i], muvalues[i]);
                    if (uncertainty)
                        unc = curves[i]->getUnintegratedEnvelope(X, ktvalues[i], muvalues[i]);
                    if (band)
                        ban = curves[i]->getUnintegratedBand(X, ktvalues[i], muvalues[i]);
                    if (mc)
                        mcban = curves[i]->getUnintegratedMC(X, ktvalues[i], muvalues[i]);
                    break;
                case TMDKT:
                    v_Y.push_back(curves[i]->getUnintegrated(xBvalues[i], X, muvalues[i]));
                    if (allmembers)
                        all = curves[i]->getUnintegratedAllMembers(xBvalues[i], X, muvalues[i]);
                    if (uncertainty)
                        unc = curves[i]->getUnintegratedEnvelope(xBvalues[i], X, muvalues[i]);
                    if (band)
                        ban = curves[i]->getUnintegratedBand(xBvalues[i], X, muvalues[i]);
                    if (mc)
                        mcban = curves[i]->getUnintegratedMC(xBvalues[i], X, muvalues[i]);
                    break;
                case TMDMU:
                    v_Y.push_back(curves[i]->getUnintegrated(xBvalues[i], ktvalues[i], X));
                    if (allmembers)
                        all = curves[i]->getUnintegratedAllMembers(xBvalues[i], ktvalues[i], X);
                    if (uncertainty)
                        unc = curves[i]->getUnintegratedEnvelope(xBvalues[i], ktvalues[i], X);
                    if (band)
                        ban = curves[i]->getUnintegratedBand(xBvalues[i], ktvalues[i], X);
                    if (mc)
                        mcban = curves[i]->getUnintegratedMC(xBvalues[i], ktvalues[i], X);
                    break;
                case PDFX:
                    v_Y.push_back(curves[i]->getIntegrated(X, muvalues[i]));
                    if (allmembers)
                        all = curves[i]->getIntegratedAllMembers(X, muvalues[i]);
                    if (uncertainty)
                        unc = curves[i]->getIntegratedEnvelope(X, muvalues[i]);
                    if (band)
                        ban = curves[i]->getIntegratedBand(X, muvalues[i]);
                    if (mc)
                        mcban = curves[i]->getIntegratedMC(X, muvalues[i]);
                    break;
                case PDFMU:
                    v_Y.push_back(curves[i]->getIntegrated(xBvalues[i], X));
                    if (allmembers)
                        all = curves[i]->getIntegratedAllMembers(xBvalues[i], X);
                    if (uncertainty)
                        unc = curves[i]->getIntegratedEnvelope(xBvalues[i], X);
                    if (band)
                        ban = curves[i]->getIntegratedBand(xBvalues[i], X);
                    if (mc)
                        mcban = curves[i]->getIntegratedMC(xBvalues[i], X);
                    break;
                default:
                    throw Notimplemented(string(type + " is not implemented."), __LINE__);
                }
                if (allmembers)
                {
                    v_all.push_back(all);
                }
                if (uncertainty)
                {
                    v_Yup .push_back(unc.first );
                    v_Ylow.push_back(unc.second);
                }
                if (band)
                {
                    v_Yup2 .push_back(ban.first );
                    v_Ylow2.push_back(ban.second);
                }
                if (mc)
                {
                    v_Yup2 .push_back(mcban.first );
                    v_Ylow2.push_back(mcban.second);
                }
            } // end inner try
            catch(UndefinedPDF& p)
            {
                cout << program_name << '\t' << p.what() << "\nIgnoring current entry." << endl;
                v_Y.push_back(vector<double>(X.size(),0.)); // add empty vector instead to make sure ncurves is still exact!
                if (allmembers)
                {
                    v_all.push_back({vector<double>(X.size(),0.)});
                }
                if (uncertainty)
                {
                    v_Yup .push_back(vector<double>(X.size(),0.));
                    v_Ylow.push_back(vector<double>(X.size(),0.));
                }
                if (band)
                {
                    v_Yup2 .push_back(vector<double>(X.size(),0.));
                    v_Ylow2.push_back(vector<double>(X.size(),0.));
                }
                if (mc)
                {
                    v_Yup2 .push_back(vector<double>(X.size(),0.));
                    v_Ylow2.push_back(vector<double>(X.size(),0.));
                }
                continue;
            }
            catch(Notimplemented& n)
            {   // catch TMDexceptions for unimplemented functions
                cout << program_name << '\t' << n.what() << "\nProgram will stop now!" << endl;
                for(unsigned int i = 0; i < ncurves; i++)
                    //delete curves[i]; // free memory
                    return EXIT_FAILURE;
            }
        } // end of for-loop
    } // end outer try
    catch(LHAPDF::RangeError l)
    {   // catch LHAPDF exceptions for unphysical x/mu/...
        cout << program_name << '\t' << l.what() << "\nProgram will stop now!" << endl;
        for(unsigned int i = 0; i < ncurves; i++)
            //delete curves[i]; // free memory
            return EXIT_FAILURE;
    }

    // Apply "pink book" draw style if selected (plot  1/s dL/dtau(sqrt(x*s)))
    //if(overs)
    //	for (unsigned int i = 0; i<X.size(); i++)
    //		for (unsigned int j = 0; j<ncurves; j++)
    //            //X[i]=sqrt(X[i]*sqrtsvalues[j]);
    //			//Y[j][i]=(0.389*Y[j][i]/(sqrtsvalues[j]*X[i])); // TODO: check this hardcoded value
    //Y[j][i] /= sqrtsvalues[j];

    // Check, if graph is inside desired y range
    vector<bool> valid_PDFplots;
    for(unsigned int i = 0; i<ncurves; i++)
    {
        valid_PDFplots.push_back(false);
        for(unsigned int j=0; j<X.size(); j++)
            valid_PDFplots[i] = (valid_PDFplots[i] || (v_Y[i][j]>ymin && v_Y[i][j]<ymax));
    }

    if (valid_PDFplots.size() == 0)
    {
        cout << program_name << "\tNo plot was produced." << endl;
        return EXIT_SUCCESS;
    }

    /*
     * Graphical stuff begins here
     */

    // First, make TGraph objects out of x and y values
    TGraph * nominal_values[ncurves],
           * upper_values[ncurves],
           * lower_values[ncurves];
//    vector<TGraph *> allmember_values[ncurves];
//    vector<TGraph *> allmember_values[100];
    vector<TGraph *> * allmember_values = new vector<TGraph *>[ncurves];

    TGraphAsymmErrors * uncertainty_bands[ncurves];
    for (unsigned int iPDF = 0 ; iPDF < ncurves ; iPDF++)
    {
        if (allmembers)
        {
            vector<TGraph *> allmember_value;
            for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                allmember_value.push_back(new TGraph((int) X.size(), &X[0], &v_all[iPDF][imember][0]));
            allmember_values[iPDF] = allmember_value;
        }

        if (uncertainty)
        {
            upper_values[iPDF] = new TGraph((int) X.size(), &X[0], &v_Yup [iPDF][0]);
            lower_values[iPDF] = new TGraph((int) X.size(), &X[0], &v_Ylow[iPDF][0]);
        }

        if (band || mc)
        {
            vector<double> exl, exh, eyl, eyh;
            for (size_t ipoint = 0 ; ipoint < X.size() ; ipoint++)
            {
                double current_Y = v_Y[iPDF][ipoint],
                       current_Yup  = v_Yup2 [iPDF][ipoint],
                       current_Ylow = v_Ylow2[iPDF][ipoint];
                exl.push_back(0); // TODO
                exh.push_back(0); // TODO
                eyl.push_back(current_Y   - current_Ylow);
                eyh.push_back(current_Yup - current_Y   );
            }
            // TGraphAsymmErrors (Int_t n, const Float_t *x, const Float_t *y, const Float_t *exl=0, const Float_t *exh=0, const Float_t *eyl=0, const Float_t *eyh=0)
            uncertainty_bands[iPDF] = new TGraphAsymmErrors((int) X.size(), &X[0], &v_Y[iPDF][0],
                    &exl[0], &exh[0], &eyl[0], &eyh[0]);
        }

        nominal_values[iPDF] = new TGraph((int) X.size(), &X[0], &v_Y[iPDF][0]);
        //nominal_values[iPDF] = new TGraph((int) X.size(), &X[0], &v_Y[iPDF][iPDF]);

        // cout << program_name << "\tContent of the graph for PDF entry " << iPDF << ":" << endl;
        // nominal_values[iPDF]->Print();
        if (!valid_PDFplots[iPDF])
            //cout << program_name << "\tCaution: The curve is outside of the y-axis range." << endl;
            cout << " Error: " << " The curve is outside of the y-axis range.Please check validity range for selected parton" << endl;
    }

    // set canvas
    const double main_pad_height = 800,
                 canvas_width = 1000,
                 ratio_pad_height = draw_ratio ? canvas_width - main_pad_height : 0;
    TCanvas * canvas = new TCanvas("tmdplotter", "TMD Plotter", canvas_width, main_pad_height + ratio_pad_height); // 3508x2480 = a4paper in landscape layout
    canvas->Draw();

    // set main pad
    // TPad (const char *name, const char *title, Double_t xlow, Double_t ylow, Double_t xup, Double_t yup, Color_t color=-1, Short_t bordersize=-1, Short_t bordermode=-2)
    TPad * main_pad = new TPad("main_pad", "TMD Plotter", 0, ratio_pad_height/canvas_width, 1, 1);
    if (X[0] > 0 && logx) main_pad->SetLogx();
    if (      ymin > 0 && logy) main_pad->SetLogy();
    main_pad->SetLeftMargin  (0.15);
    main_pad->SetRightMargin (0.10);
    main_pad->SetBottomMargin(0.11);
    main_pad->SetTopMargin   (0.09);
    canvas->cd();
    main_pad->Draw();
    main_pad->SetTicks();
    main_pad->cd();

    // set frame
    TH1F * frame = new TH1F("main_frame","",nsteps,X.front(),X.back());
    frame->SetMinimum(ymin);
    frame->SetMaximum(ymax);
    frame->SetStats(0);
    frame->SetTitleOffset(1.0,"X");
    frame->SetTitleOffset(1.4,"Y");

    // set title
    switch(type)
    {
    case TMDKT:
        if (mode == LUMINOSITY)
            frame->SetXTitle("q_{t} [GeV]");
        else
            frame->SetXTitle("k_{t} [GeV]");
        break;
    case TMDMU:
    case PDFMU:
        if (mode == LUMINOSITY) // TODO?
            frame->SetXTitle(overs ? "#sqrt{#hat{s}} [TeV]" : "#tau");
        else
            frame->SetXTitle("p [GeV]");
        break;
    case TMDX:
    case PDFX:
        if (mode == LUMINOSITY)
            frame->SetXTitle(overs ? "#sqrt{#hat{s}} [TeV]" : "#tau");
        else
            frame->SetXTitle("x");
    }
    switch(mode)
    {
    case DEFAULT:
        switch (type)
        {
        case TMDX:
        case TMDMU:
        case TMDKT:
//                   frame->SetYTitle("xA(x,k_{t},p)");
            frame->SetYTitle("xA(x,k_{t},#mu)");
            break;
        case PDFX:
        case PDFMU:
//                    frame->SetYTitle("xf(x,p)");
            frame->SetYTitle("xf(x,#mu)");
        }
        break;
    case LUMINOSITY:
    {
        string ytitle = overs ? "#frac{1}{s} " : "";
        switch (type)
        {
        case TMDX:
        case TMDKT:
//                    ytitle += "#frac{d^{2}L}{d#taudq_{t}}(#tau,q_{t},p) [GeV^{-1}]";
            ytitle += "#frac{d^{2}L}{d#taudq_{t}}(#tau,q_{t},#mu) [GeV^{-1}]";
            break;
        case PDFX:
        case PDFMU:
//                    ytitle += "#frac{dL}{d#tau}(#tau,p)";
            ytitle += "#frac{dL}{d#tau}(#tau,#mu)";
        }
        ytitle += overs ? " [pb]" : ""; // TODO: check unit
        frame->SetYTitle(ytitle.c_str());
        break;
    }
    } // end mode switch, now set label size
    const double label_size = 0.04; // used
    frame->SetLabelSize(label_size, "XY");
    frame->SetTitleSize(1.2*label_size, "XY");

    // graph style
    vector<Color_t> colors = {kRed, kBlue, kMagenta+2, kOrange+7, kGreen+3, kCyan+2, kAzure+3};
    const unsigned Ncolors = colors.size();
    for (unsigned short iPDF = 0 ; iPDF < ncurves ; iPDF++)
    {
        Color_t current_color = colors[iPDF%Ncolors];
        nominal_values[iPDF]->SetLineColor(current_color);
        nominal_values[iPDF]->SetLineStyle(1*iPDF/Ncolors);
        nominal_values[iPDF]->SetLineWidth(3);
        if (allmembers)
        {
            for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
            {
                allmember_values[iPDF][imember]->SetLineColor(current_color);
                allmember_values[iPDF][imember]->SetLineWidth(1);
                //if (central_value) allmember_values[iPDF][imember]->SetLineStyle(3);
            }
        }

        if (uncertainty)
        {
            upper_values[iPDF]->SetLineColor(current_color);
            lower_values[iPDF]->SetLineColor(current_color);
            upper_values[iPDF]->SetLineWidth(2);
            lower_values[iPDF]->SetLineWidth(2);
            //if (central_value) upper_values[iPDF]->SetLineStyle(9);
            //if (central_value) lower_values[iPDF]->SetLineStyle(9);
        }

        if (band)
        {
            uncertainty_bands[iPDF]->SetLineColorAlpha(current_color, 0.2);
            uncertainty_bands[iPDF]->SetFillColorAlpha(current_color, 0.2);
        }
        if (mc)
        {
            uncertainty_bands[iPDF]->SetLineColorAlpha(current_color, 0.2);
            uncertainty_bands[iPDF]->SetFillColorAlpha(current_color, 0.2);
        }
    }

    // legend
    // TLegend(Double_t x1, Double_t y1, Double_t x2, Double_t y2, const char* header = "", Option_t* option = "brNDC")
    const double legend_text_size = 0.03,
                 legend_ymax = 0.88;
    TLegend * mylegend = new TLegend(0.2,legend_ymax - ncurves*legend_text_size,0.8,legend_ymax);
    mylegend->SetFillStyle(0); // transparent background
    mylegend->SetBorderSize(0); // get rid of the borders
    mylegend->SetTextSize(legend_text_size);
    if (ncurves > 0)
    {
        // try to put in title what is common in the legend to all curves
        // -> initialise the common value by the first element, and as soon as sth different is found, set it to NULL
        map<const string,string> elements = {{"evolution", curves[0]->getEvolution  ()},
            {"parton", curves[0]->getParton     ()},
            {"set", curves[0]->getSet        ()},
            {"sf", remove_trailing_zeros(to_string(curves[0]->getScalefactor()))}
        };

        if (type == TMDKT)
        {
            elements["x"] = to_string(xBvalues[0]);
            elements["p"] = to_string(muvalues[0]);
        }
        else if (type == TMDX)
        {
            elements["kt"] = to_string(ktvalues[0]);
            elements["p"  ] = to_string(muvalues[0]);
        }
        else if (type == TMDMU)
        {
            elements["kt"] = to_string(ktvalues[0]);
            elements["x"  ] = to_string(xBvalues[0]);
        }
        else if (type == PDFX)
        {
            elements["p"]        = to_string(muvalues[0]);
            elements["intktmin"] = to_string(curves[0]->getKtMin());
            elements["intktmax"] = to_string(curves[0]->getKtMax());
        }
        else if (type == PDFMU)
        {
            elements["x"]        = to_string(xBvalues[0]);
            elements["intktmin"] = to_string(curves[0]->getKtMin());
            elements["intktmax"] = to_string(curves[0]->getKtMax());
        }

        if (mode == LUMINOSITY)
            elements["s"] = to_string(sqrtsvalues[0]);

        // check common values
        for (size_t icurve = 1 ; icurve < ncurves ; icurve++)
        {
            // check parton
            if (elements["evolution"] != curves[icurve]->getEvolution  ()) elements["evolution"].clear();
            if (elements["parton"   ] != curves[icurve]->getParton     ()) elements["parton"   ].clear();
            if (elements["set"      ] != curves[icurve]->getSet        ()) elements["set"      ].clear();
            if (elements["sf"       ] != remove_trailing_zeros(to_string(curves[icurve]->getScalefactor()))) elements["sf"].clear();
            // check parameters (if existing)
            if (elements.find("x"        ) != elements.end() && elements["x"       ] != to_string(xBvalues[icurve]          )) elements["x"].clear();
            if (elements.find("p"        ) != elements.end() && elements["p"       ] != to_string(muvalues[icurve]          )) elements["p"].clear();
            if (elements.find("s"        ) != elements.end() && elements["s"       ] != to_string(sqrtsvalues[icurve]       )) elements["s"].clear();
            if (elements.find("kt"       ) != elements.end() && elements["kt"      ] != to_string(ktvalues[icurve]          )) elements["kt"].clear();
            if (elements.find("intktmin" ) != elements.end() && elements["intktmin"] != to_string(curves[icurve]->getKtMin())) elements["intktmin"].clear();
            if (elements.find("intktmax" ) != elements.end() && elements["intktmax"] != to_string(curves[icurve]->getKtMax())) elements["intktmax"].clear();
        }

        const bool find_x       = (elements.find("x"       ) != elements.end()),
                   find_p       = (elements.find("p"       ) != elements.end()),
                   find_s       = (elements.find("s"       ) != elements.end()),
                   find_kt      = (elements.find("kt"      ) != elements.end()),
                   find_ktrange = (elements.find("intktmin") != elements.end())
                                  && (elements.find("intktmax") != elements.end());
        const bool //common_evolution = (elements.find("evolution") != elements.end() && !elements["evolution"].empty()), // TODO: subdivide the legend into the TMD and DGLAP curves
        common_parton  = (!elements["parton"].empty()),
        common_set     = (!elements["set"   ].empty()),
        common_sf      = (!elements["sf"    ].empty()),
        common_x       = (!elements["x"     ].empty()),
        common_p       = (!elements["p"     ].empty()) && ((elements["evolution"] == "TMD") || type == PDFX || type == PDFMU), // TODO: check... shouldn't this be for kt??
        common_s       = (!elements["s"     ].empty()) && overs,
        common_kt      = (!elements["kt"    ].empty()),
        common_ktrange = (elements["evolution"] == "TMD")
                         && (!elements["intktmin"].empty())
                         && (!elements["intktmax"].empty());
        string main_title;
        if (                common_parton ) main_title += elements["parton"] + ", ";
        if (                common_set    ) main_title += elements["set"   ] + ", ";
        if (find_x       && common_x      ) main_title += (mode == LUMINOSITY ? "#tau = " : "x = ") + remove_trailing_zeros(elements["x" ]) + ", ";
//        if (find_p       && common_p      ) main_title += "p = "        + remove_trailing_zeros(elements["p" ]) + " GeV, ";
        if (find_p       && common_p      ) main_title += "#mu = "        + remove_trailing_zeros(elements["p" ]) + " GeV, ";
        if (find_s       && common_s      ) main_title += "s = "        + remove_trailing_zeros(elements["s" ]) + " TeV, ";
        if (find_kt      && common_kt     ) main_title += (mode == LUMINOSITY ? "q_{t} = " : "k_{t} = ") + remove_trailing_zeros(elements["kt"]) + " GeV, ";
        if (find_ktrange && common_ktrange) main_title += "k_{t} from " + remove_trailing_zeros(elements["intktmin"]) + " up to " + remove_trailing_zeros(elements["intktmax"]) + " GeV, ";

        if (main_title.length() > 2) main_title.erase(main_title.end()-2,main_title.end()); // remove the last ", " if applicable
        if (common_sf && curves[0]->getScalefactor() != 1) main_title += " #times " + elements["sf"];
        frame->SetTitle(main_title.c_str());

        for (size_t icurve = 0 ; icurve < ncurves ; icurve++)
        {
            string label;
            if (                !common_parton) label += curves[icurve]->getParton() + ", ";
            if (                !common_set   ) label += curves[icurve]->getSet   () + ", ";
            if (find_x       && !common_x) label += (mode == LUMINOSITY ? "#tau = " : "x = ") + remove_trailing_zeros(to_string(xBvalues [icurve])) + ", ";
//            if (find_p       && !common_p && (curves[icurve]->getEvolution() == "TMD" || type == PDFX || type == PDFMU)) label += "p = " + remove_trailing_zeros(to_string(muvalues [icurve])) + " GeV, ";
            if (find_p       && !common_p && (curves[icurve]->getEvolution() == "TMD" || type == PDFX || type == PDFMU)) label += "#mu = " + remove_trailing_zeros(to_string(muvalues [icurve])) + " GeV, ";
            if (find_s       && !common_s && overs) label += "#sqrt{s}= " + remove_trailing_zeros(to_string(sqrtsvalues[icurve])) + " TeV, ";
            if (find_kt      && !common_kt) label += (mode == LUMINOSITY ? "q_{t} = " : "k_{t} = ") + remove_trailing_zeros(to_string(ktvalues[icurve])) + " GeV, ";
            if (find_ktrange && !common_ktrange && curves[icurve]->getEvolution() == "TMD") label += "k_{t} from " + remove_trailing_zeros(to_string(curves[icurve]->getKtMin())) + " up to " + remove_trailing_zeros(to_string(curves[icurve]->getKtMax())) + " GeV, ";

            if (label.length() > 2) label.erase(label.end()-2,label.end()); // remove the last ", " if applicable
            if (!common_sf && (curves[icurve]->getScalefactor() != 1)) label += " #times " + remove_trailing_zeros(to_string(curves[icurve]->getScalefactor()));
            mylegend->AddEntry(nominal_values[icurve], label.c_str(), "l");
        }
    }
    frame->Draw();
    if (band)
        for (unsigned short iPDF = 0 ; iPDF < ncurves ; iPDF++)
            if (valid_PDFplots[iPDF]) {
                //uncertainty_bands[iPDF]->Draw("03same");
                uncertainty_bands[iPDF]->Draw("03Csame");
            }
    for (unsigned short iPDF = 0 ; iPDF < ncurves ; iPDF++)
        if (valid_PDFplots[iPDF])
        {
            if (allmembers)
                for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                    allmember_values[iPDF][imember]->Draw("Csame");

            if (uncertainty)
            {
                upper_values[iPDF]->Draw("Csame");
                lower_values[iPDF]->Draw("Csame");
            }
            if (central_value)
//                nominal_values[iPDF]->Draw("Csame");
                nominal_values[iPDF]->Draw("Lsame");
        }
    if (ncurves > 1)
        mylegend->Draw();
    main_pad->RedrawAxis();

    // ratio drawing stuff
    if (draw_ratio)
    {
        // TPad (const char *name, const char *title, Double_t xlow, Double_t ylow, Double_t xup, Double_t yup, Color_t color=-1, Short_t bordersize=-1, Short_t bordermode=-2)
        canvas->cd();
        TPad * ratio_pad = new TPad("ratio_pad", "", 0, 0, 1, 1.35*ratio_pad_height/canvas_width);
        if (X.front() > 0 && logx) ratio_pad->SetLogx();
        if (ratioymin > 0 && ratiology) ratio_pad->SetLogy();
        ratio_pad->SetLeftMargin  (0.15);
        ratio_pad->SetRightMargin (0.10);
//		ratio_pad->SetBottomMargin(0.16);
        ratio_pad->SetBottomMargin(0.3);
//		ratio_pad->SetTopMargin   (0.04);
        ratio_pad->SetTopMargin   (0.04);
        ratio_pad->SetTicks();
        ratio_pad->Draw();
        ratio_pad->cd();

        TGraph * ratio_nominal[ncurves],
               * ratio_upper  [ncurves],
               * ratio_lower  [ncurves];
//        vector<TGraph *> ratio_allmembers[ncurves];
//        vector<TGraph *> ratio_allmembers[100];
        vector<TGraph *> * ratio_allmembers = new vector<TGraph *>[ncurves];

        TGraphAsymmErrors * ratio_uncertainty_bands[ncurves];
        for (size_t iPDF = 0 ; iPDF < ncurves ; iPDF++)
        {
            if (allmembers)
                for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                    ratio_allmembers[iPDF].push_back(static_cast<TGraph *>(allmember_values[iPDF][imember]->Clone()));
            if (band)
                ratio_uncertainty_bands[iPDF] = static_cast<TGraphAsymmErrors *>(uncertainty_bands[iPDF]->Clone());
            if (mc)
                ratio_uncertainty_bands[iPDF] = static_cast<TGraphAsymmErrors *>(uncertainty_bands[iPDF]->Clone());
            if (uncertainty)
            {
                ratio_upper[iPDF] = static_cast<TGraph *>(upper_values[iPDF]->Clone());
                ratio_lower[iPDF] = static_cast<TGraph *>(lower_values[iPDF]->Clone());
            }
            ratio_nominal[iPDF] = static_cast<TGraph *>(nominal_values[iPDF]->Clone());
        }
        for (size_t istep = 0 ; istep <= nsteps ; istep++)
        {
            double x_den, y_den;
            nominal_values[0]->GetPoint(istep, x_den, y_den);
            if (y_den == 0) // stop to avoid NaNs from ratios
            {
                for (unsigned short iPDF = 0 ; iPDF < ncurves ; iPDF++)
                {
                    //nominal_values[iPDF]->GetPoint(istep, x_den, y_den);
                    ratio_nominal[iPDF]->Set(max(0,int(istep-1)));
                    if (allmembers)
                        for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                            ratio_allmembers[iPDF][imember]->Set(max(0,int(istep-1)));
                    if (band)
                        ratio_uncertainty_bands[iPDF]->Set(max(0,int(istep-1)));
                    if (mc)
                        ratio_uncertainty_bands[iPDF]->Set(max(0,int(istep-1)));
                    if (uncertainty)
                    {
                        ratio_upper[iPDF]->Set(max(0,int(istep-1)));
                        ratio_lower[iPDF]->Set(max(0,int(istep-1)));
                    }
                }
                break;
            }
            else
            {
                for (unsigned short iPDF = 0 ; iPDF < ncurves ; iPDF++)
                {
                    double x_num, y_num;
                    nominal_values[iPDF]->GetPoint(istep, x_num, y_num);
                    // cout << " calculating ratio " << iPDF << " " << istep << endl;
                    if(istep < nsteps ) {
                        ratio_nominal[iPDF]->SetPoint(istep, x_num, y_num/y_den);
                    }
                    else {ratio_nominal[iPDF]->SetPoint(istep, x_num, 1.);}
                    if (allmembers)
                    {
                        for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                        {
                            allmember_values[iPDF][imember]->GetPoint(istep, x_num, y_num);
                            ratio_allmembers[iPDF][imember]->SetPoint(istep, x_num, y_num/y_den);
                        }
                    }
                    if (band || uncertainty || mc)
                    {
                        double current_x_den = 0, current_y_den = 0;
                        nominal_values[iPDF]->GetPoint(istep, current_x_den, current_y_den);
                        if (band)
                        {
                            if(current_y_den == 0 || current_y_den < ymin ) { //cout << "current den = 0 " << endl;
                            }
                            else {
                                ratio_uncertainty_bands[iPDF]->GetPoint(istep, x_num, y_num);
                                ratio_uncertainty_bands[iPDF]->SetPoint(istep, x_num, y_num/current_y_den);
                                ratio_uncertainty_bands[iPDF]->SetPointEYhigh(istep, uncertainty_bands[iPDF]->GetErrorYhigh(istep)/current_y_den);
                                ratio_uncertainty_bands[iPDF]->SetPointEYlow (istep, uncertainty_bands[iPDF]->GetErrorYlow (istep)/current_y_den);
                            }
                        }
                        if (mc)
                        {
                            if(current_y_den == 0 || current_y_den < ymin ) { //cout << "current den = 0 " << endl;
                            } else {
                                ratio_uncertainty_bands[iPDF]->GetPoint(istep, x_num, y_num);
                                ratio_uncertainty_bands[iPDF]->SetPoint(istep, x_num, y_num/current_y_den);
                                ratio_uncertainty_bands[iPDF]->SetPointEYhigh(istep, uncertainty_bands[iPDF]->GetErrorYhigh(istep)/current_y_den);
                                ratio_uncertainty_bands[iPDF]->SetPointEYlow (istep, uncertainty_bands[iPDF]->GetErrorYlow (istep)/current_y_den);
                            }
                        }
                        if (uncertainty)
                        {
                            if(current_y_den == 0 || current_y_den < ymin ) { //cout << "current den = 0 " << endl;
                            }
                            else {
                                upper_values[iPDF]->GetPoint(istep, x_num, y_num);
                                ratio_upper[iPDF]->SetPoint(istep, x_num, y_num/current_y_den);
                                lower_values[iPDF]->GetPoint(istep, x_num, y_num);
                                ratio_lower[iPDF]->SetPoint(istep, x_num, y_num/current_y_den);
                            }
                        }
                    }
                }
            }
        }
        TH1F * ratio_frame = new TH1F("ratio_frame","",nsteps,X.front(),X.back());
        ratio_frame->SetMinimum(ratioymin);
        ratio_frame->SetMaximum(ratioymax);
        ratio_frame->SetStats(0);
        ratio_frame->SetTitleOffset(1.3,"X");
        ratio_frame->SetTitleOffset(1.4,"Y");
        const double ratio_label_size = 0.6*label_size/(ratio_pad_height/canvas_width);
        frame->SetLabelSize(0, "X");
        frame->SetTitleSize(0, "X");
        ratio_frame->SetXTitle(frame->GetXaxis()->GetTitle());
        ratio_frame->SetLabelSize(ratio_label_size,"X");
        ratio_frame->SetLabelSize(ratio_label_size,"Y");
        ratio_frame->SetTitleSize(ratio_label_size,"X");
        ratio_frame->SetTitleSize(ratio_label_size,"Y");
        ratio_frame->SetNdivisions(304, "Y");
        ratio_frame->SetTickLength (0.1, "X");
        ratio_frame->Draw();
        if (band || mc )
            for (size_t iPDF = 0 ; iPDF < ncurves ; iPDF++)
                if (valid_PDFplots[iPDF]) {
                    //ratio_uncertainty_bands[iPDF]->SetLineWidth(1504);
                    ratio_uncertainty_bands[iPDF]->Draw("03same");
                    //ratio_uncertainty_bands[iPDF]->Draw("Csame");
                }
        for (size_t iPDF = 0 ; iPDF < ncurves ; iPDF++)
        {
            // cout << program_name << "\tratio " << iPDF << ":" << endl;
            // ratio_nominal[iPDF]->Print();
            if (valid_PDFplots[iPDF])
            {
                // cout << " test band " << band << endl;
                if (central_value && !band)
                    ratio_nominal[iPDF]->Draw("Csame");
                //ratio_nominal[iPDF]->SetLineColor(kBlack);
                else if (iPDF == 0 && !band)
                {
                    ratio_nominal[iPDF]->SetLineColor(kBlack);
                    ratio_nominal[iPDF]->SetLineWidth(2);
                    ratio_nominal[iPDF]->Draw("Csame");
                }

                if (allmembers)
                    for (unsigned int imember = 0 ; imember < v_all[iPDF].size() ; imember++)
                        ratio_allmembers[iPDF][imember]->Draw("Csame");
                if (uncertainty)
                {
                    ratio_upper[iPDF]->SetFillColorAlpha(0,0);
                    ratio_upper[iPDF]->SetLineWidth(-1401);
                    ratio_upper[iPDF]->Draw("Csame");
                    ratio_lower[iPDF]->SetLineWidth(-1401);
                    ratio_lower[iPDF]->SetFillColorAlpha(0,0);
                    ratio_lower[iPDF]->Draw("Csame");
                }
            }
        }
        ratio_pad->RedrawAxis();
    }

    // logo
    main_pad->cd();
    //TPad * pad_logo = new TPad ("logo", "logo", 0.5,legend_ymax,0.8,0.8,2);
    TPad * pad_logo = new TPad ("logo", "logo", 0.91,0.1,0.96,0.9);
    pad_logo->Draw();
    pad_logo->cd();
    ////TASImage * logo = new TASImage ("/www/html/include/cookie.png", kPng);
    ////logo->Draw("X");
    TLatex * logo_text = new TLatex(0.5,0.2,"TMDplotter 2.2.4");
    logo_text->SetTextAngle(90);
    logo_text->SetTextColor();
    logo_text->SetTextAlign(22);
    logo_text->SetTextFont(43);
    logo_text->SetTextSize(30);
    logo_text->Draw();
    // cout << " outfile " << output_name.c_str() <<  " " <<filetype.c_str() << endl;
    canvas->Print(output_name.c_str(), filetype.c_str());
    if (filetype != "png")
        canvas->Print((directory + filename + string(".png")).c_str(), "png");
    //canvas->SaveAs((directory + filename + string(".png")).c_str());
    // free memory
    for(unsigned int i=0; i<ncurves; i++) {
        //delete curves[i];
        delete nominal_values[i];
    }
    int stop = time(NULL);
    cout << " Info: Time needed for plotting: " << (stop-start) << " sec " << endl;
    return EXIT_SUCCESS;
}
