//#define DEBUG

#include "tmdlib/TMDlib.h"



using namespace TMDlib;

/***
* Routine to interpolate Pavia TMDgrids and to return momentum weighted TMD parton density as
* function of x, kt, scale p for different flavors.
***/


#include "tmdlib/factories.h"

#include <iostream>
#include <sstream>
#include <string>
#include <fstream>


using namespace TMDlib;

/* Common Block Declarations */
extern std::string pdfpath;

/***
* Routine to interpolate Pavia TMDgrids and to return momentum weighted TMD parton density as
* function of x, kt, scale p for different flavors.
***/

void TMD::Pavia(double x, double kt, double p, double& up, double& ubar, double& dn, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& top, double& tbar, double& phot, double& glu)
{
    if (first) {
        // Define file names
        // if (TMD_Mem == "0010" || TMD_Mem == "0034" || TMD_Mem == "0039" || TMD_Mem == "0041" || TMD_Mem == "0043") TMD_Mem = "0009" ;
        std::string grid = pdfpath + TMD_Dir + TMD_Name + "_" + TMD_Mem + ".dat";
        std::string info = pdfpath + TMD_Dir + TMD_Name + ".info";
        if ( TMDverbosity() == 2) {
            std::cout << " [Pavia]: file = " << grid << std::endl;
            std::cout << " [Pavia]: Member = " << TMD_Mem << std::endl;
        }

        // Read info file
        // if ( TMDverbosity() == 2) std::cout << "Reading info file: " + info +  "  ..." << std::std::endl;
        TMDLIB_YAML::Node const& infofile = TMDLIB_YAML::LoadFile(info);

        // Read type of TMD distribution
        const std::string pf = infofile["TMDType"].as<std::string>();

        // Read TMDs in grid
        TMDs = TMDlib::mkTMD(TMD_Name, pdfpath, std::stoi(TMD_Mem));
        first = false;
    }
    // Values of xTMD(x,k_T,Q) for the different flavours.
    // The interpolator works as TMDs->Evaluate(x , kToQ * Q , Q).at(ifl)
    const std::map<int, double> PaviaTMDs = TMDs->Evaluate(x, kt, p);

    glu     = 0; // not in the grids
    up      = PaviaTMDs.at(2);
    ubar    = PaviaTMDs.at(-2);
    dn      = PaviaTMDs.at(1);
    dbar    = PaviaTMDs.at(-1);
    strange = PaviaTMDs.at(3);
    sbar    = PaviaTMDs.at(-3);
    charm   = PaviaTMDs.at(4);
    cbar    = PaviaTMDs.at(-4);
    bottom  = PaviaTMDs.at(5);
    bbar    = PaviaTMDs.at(-5);
    top     = 0; // not in the grids
    tbar    = 0; // not in the grids
    phot    = 0; // not in the grids

    // Check
    // std::cout << " [Pavia]:  up  = " << up << "  -*-  kt = "  << " " << kt << std::std::endl;

    // delete TMDs;
    return ;
}

//
// Authors: Valerio Bertone: valerio.bertone@cern.ch
//          Chiara Bissolotti: chiara.bissolotti01@universitadipavia.it
//

#include "tmdlib/structgrid.h"

namespace TMDlib
{
//_________________________________________________________________________________
StructGrid::StructGrid(TMDLIB_YAML::Node const& info, TMDLIB_YAML::Node const& grid):
    _info(info),
    _xg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double>
{
    grid["xg"].as<std::vector<double>>(), 3
})),
_zg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double>
{
    grid["zg"].as<std::vector<double>>(), 3
})),
_qToQg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double> {grid["qToQg"].as<std::vector<double>>(), 3})),
_Qg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double> {grid["Qg"].as<std::vector<double>>(), 3})),
_stfunc(grid["StructureFunction"].as<std::vector<std::vector<std::vector<std::vector<double>>>>>())
{
}

//_________________________________________________________________________________
double StructGrid::Evaluate(double const& x, double const& z, double const& qT, double const& Q) const
{
    // If qT/Q < the first point of the grid, put qT/Q  equal to it.
    // Do not compute below the first point of the grid.
    const double qToQ = std::max(qT / Q, _qToQg->GetQGrid().front());


    // Get summation bounds
    const std::tuple<int, int, int> xbounds    = _xg->SumBounds(x);
    const std::tuple<int, int, int> zbounds    = _zg->SumBounds(z);
    const std::tuple<int, int, int> qToQbounds = _qToQg->SumBounds(qToQ);
    const std::tuple<int, int, int> Qbounds    = _Qg->SumBounds(Q);

    // Compute interpolating functions singularly to minimise the
    // number of calls due to the the nesting of the interpolation.
    // Q grid
    const int inQ = std::get<1>(Qbounds);
    const int nQ  = std::get<2>(Qbounds) - inQ;
    std::vector<double> IQ(nQ);
    for (int iQ = 0; iQ < nQ; iQ++)
        IQ[iQ] = _Qg->Interpolant(std::get<0>(Qbounds), inQ + iQ, Q);

    // x grid
    const int inx = std::get<1>(xbounds);
    const int nx  = std::get<2>(xbounds) - inx;
    std::vector<double> Ix(nx);
    for (int ix = 0; ix < nx; ix++)
        Ix[ix] = _xg->Interpolant(std::get<0>(xbounds), inx + ix, x);

    // z grid
    const int inz = std::get<1>(zbounds);
    const int nz  = std::get<2>(zbounds) - inz;
    std::vector<double> Iz(nz);
    for (int iz = 0; iz < nz; iz++)
        Iz[iz] = _zg->Interpolant(std::get<0>(zbounds), inz + iz, z);

    // qT grid
    const int inqT = std::get<1>(qToQbounds);
    const int nqT  = std::get<2>(qToQbounds) - inqT;
    std::vector<double> IqT(nqT);
    for (int iqT = 0; iqT < nqT; iqT++)
        IqT[iqT] = _qToQg->Interpolant(std::get<0>(qToQbounds), inqT + iqT, qToQ);

    // Initialise output to zero
    double result = 0;

    // Do the interpolation
    for (int iQ = 0; iQ < nQ; iQ++)
        for (int ix = 0; ix < nx; ix++)
            for (int iz = 0; iz < nz; iz++)
                for (int iqT = 0; iqT < nqT; iqT++)
                    result += IQ[iQ] * Ix[ix] * Iz[iz] * IqT[iqT] * _stfunc[inQ + iQ][inx + ix][inz + iz][inqT + iqT];

    return result;
}
}

//
// Author: Valerio Bertone: valerio.bertone@cern.ch
//

#include "tmdlib/factories.h"
#include "tmdlib/numtostring.h"

namespace TMDlib
{
//_________________________________________________________________________________
TMDGrid* mkTMD(std::string const& name, int const& mem)
{
    // std::cout << "[TMDlib]: loading " << name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
    return new TMDlib::TMDGrid{TMDLIB_YAML::LoadFile(name + "/" + name + ".info"),
                               TMDLIB_YAML::LoadFile(name + "/" + name + "_" + num_to_string(mem) + ".dat")};
}

//_________________________________________________________________________________
TMDGrid* mkTMD(std::string const& name, std::string const& folder, int const& mem)
{
    // std::cout << "[TMDlib]: loading " << folder + "/" + name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
    return new TMDlib::TMDGrid{TMDLIB_YAML::LoadFile(folder + "/"  + name + "/" + name + ".info"),
                               TMDLIB_YAML::LoadFile(folder + "/"  + name + "/" + name + "_" + num_to_string(mem) + ".dat")};
}

//_________________________________________________________________________________
std::vector<TMDGrid*> mkTMDs(std::string const& name)
{
    const TMDLIB_YAML::Node info = TMDLIB_YAML::LoadFile(name + "/" + name + ".info");
    const int nmem = info["NumMembers"].as<int>();
    std::vector<TMDGrid*> tmds(nmem);
    for (int mem = 0; mem < nmem; mem++)
    {
        // std::cout << "[TMDlib]: loading " << name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
        tmds[mem] = new TMDlib::TMDGrid{info, TMDLIB_YAML::LoadFile(name + "/" + name + "_" + num_to_string(mem) + ".dat")};
    }
    return tmds;
}

//_________________________________________________________________________________
StructGrid* mkSF(std::string const& name, int const& mem)
{
    //  std::cout << "[TMDlib]: loading " << name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
    return new TMDlib::StructGrid{TMDLIB_YAML::LoadFile(name + "/" + name + ".info"),
                                  TMDLIB_YAML::LoadFile(name + "/" + name + "_" + num_to_string(mem) + ".dat")};
}

//_________________________________________________________________________________
StructGrid* mkSF(std::string const& name, std::string const& folder, int const& mem)
{
    // std::cout << "[TMDlib]: loading " << folder + "/" + name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
    return new TMDlib::StructGrid{TMDLIB_YAML::LoadFile(folder + "/"  + name + "/" + name + ".info"),
                                  TMDLIB_YAML::LoadFile(folder + "/"  + name + "/" + name + "_" + num_to_string(mem) + ".dat")};
}

//_________________________________________________________________________________
std::vector<StructGrid*> mkSFs(std::string const& name)
{
    const TMDLIB_YAML::Node info = TMDLIB_YAML::LoadFile(name + "/" + name + ".info");
    const int nmem = info["NumMembers"].as<int>();
    std::vector<StructGrid*> sfs(nmem);
    for (int mem = 0; mem < nmem; mem++)
    {
        // std::cout << "[TMDlib]: loading " << name + "/" + name + "_" + num_to_string(mem) + ".dat" << std::std::endl;
        sfs[mem] = new TMDlib::StructGrid{info, TMDLIB_YAML::LoadFile(name + "/" + name + "_" + num_to_string(mem) + ".dat")};
    }

    return sfs;
}
}

//
// Authors: Valerio Bertone: valerio.bertone@cern.ch
//

#include "tmdlib/tmdgrid.h"

namespace TMDlib
{
//_________________________________________________________________________________
TMDGrid::TMDGrid(TMDLIB_YAML::Node const& info, TMDLIB_YAML::Node const& grid):
    _info(info),
    _xg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double>
{
    grid["xg"].as<std::vector<double>>(), 3
})),
_qToQg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double> {grid["qToQg"].as<std::vector<double>>(), 3})),
_Qg(std::unique_ptr<apfel::QGrid<double>>(new apfel::QGrid<double> {grid["Qg"].as<std::vector<double>>(), 3})),
_tmds(grid["TMDs"].as<std::map<int, std::vector<std::vector<std::vector<double>>>>>())
{
}

//_________________________________________________________________________________
std::map<int, double> TMDGrid::Evaluate(double const& x, double const& qT, double const& Q) const
{
    // If qT/Q < the first point of the grid, put qT/Q  equal to it.
    // Do not compute below the first point of the grid.
    const double qToQ = std::max(qT / Q, _qToQg->GetQGrid().front());

    // Get summation bounds
    const std::tuple<int, int, int> xbounds    = _xg->SumBounds(x);
    const std::tuple<int, int, int> qToQbounds = _qToQg->SumBounds(qToQ);
    const std::tuple<int, int, int> Qbounds    = _Qg->SumBounds(Q);

    // Compute interpolating functions singularly to minimise the
    // number of calls due to the the nesting of the interpolation.
    // Q grid
    const int inQ = std::get<1>(Qbounds);
    const int nQ  = std::get<2>(Qbounds) - inQ;
    std::vector<double> IQ(nQ);
    for (int iQ = 0; iQ < nQ; iQ++)
        IQ[iQ] = _Qg->Interpolant(std::get<0>(Qbounds), inQ + iQ, Q);

    // x grid
    const int inx = std::get<1>(xbounds);
    const int nx  = std::get<2>(xbounds) - inx;
    std::vector<double> Ix(nx);
    for (int ix = 0; ix < nx; ix++)
        Ix[ix] = _xg->Interpolant(std::get<0>(xbounds), inx + ix, x);

    const int inqT = std::get<1>(qToQbounds);
    const int nqT  = std::get<2>(qToQbounds) - inqT;
    std::vector<double> IqT(nqT);
    for (int iqT = 0; iqT < nqT; iqT++)
        IqT[iqT] = _qToQg->Interpolant(std::get<0>(qToQbounds), inqT + iqT, qToQ);

    // Initialise output map to zero
    std::map<int, double> result;
    for (auto const& tmd : _tmds)
        result.insert({tmd.first, 0});

    // Do the interpolation
    for (auto const& tmd : _tmds)
        for (int iQ = 0; iQ < nQ; iQ++)
            for (int ix = 0; ix < nx; ix++)
                for (int iqT = 0; iqT < nqT; iqT++)
                    result[tmd.first] += IQ[iQ] * Ix[ix] * IqT[iqT] * tmd.second[inQ + iQ][inx + ix][inqT + iqT];

    return result;
}
}

